@@ -1,5 +1,5 @@
 /* GObject - GLib Type, Object, Parameter and Signal Library
- * Copyright (C) 2000 Red Hat, Inc.
+ * Copyright (C) 2000-2001 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -16,6 +16,7 @@
  * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
  * Boston, MA 02111-1307, USA.
  */
+
 #include	"gclosure.h"
 
 #include	"gvalue.h"
@@ -157,7 +158,7 @@ g_closure_set_meta_marshal (GClosure       *closure,
   g_return_if_fail (meta_marshal != NULL);
   g_return_if_fail (closure->is_invalid == FALSE);
   g_return_if_fail (closure->in_marshal == FALSE);
-  g_return_if_fail (closure->meta_marshal == FALSE);
+  g_return_if_fail (closure->meta_marshal == 0);
 
   n = CLOSURE_N_NOTIFIERS (closure);
   notifiers = closure->notifiers;
@@ -220,9 +221,9 @@ g_closure_add_marshal_guards (GClosure      *closure,
 }
 
 void
-g_closure_add_fnotify (GClosure      *closure,
-		       gpointer       notify_data,
-		       GClosureNotify notify_func)
+g_closure_add_finalize_notifier (GClosure      *closure,
+				 gpointer       notify_data,
+				 GClosureNotify notify_func)
 {
   guint i;
 
@@ -242,9 +243,9 @@ g_closure_add_fnotify (GClosure      *closure,
 }
 
 void
-g_closure_add_inotify (GClosure      *closure,
-		       gpointer       notify_data,
-		       GClosureNotify notify_func)
+g_closure_add_invalidate_notifier (GClosure      *closure,
+				   gpointer       notify_data,
+				   GClosureNotify notify_func)
 {
   guint i;
 
@@ -310,14 +311,7 @@ g_closure_ref (GClosure *closure)
   g_return_val_if_fail (closure->ref_count > 0, NULL);
   g_return_val_if_fail (closure->ref_count < CLOSURE_MAX_REF_COUNT, NULL);
 
-  /* floating is basically a kludge to avoid creating closures
-   * with a ref_count of 0. so the first one doing _ref() will
-   * own the closure's initial ref_count
-   */
-  if (closure->floating)
-    closure->floating = FALSE;
-  else
-    closure->ref_count += 1;
+  closure->ref_count += 1;
 
   return closure;
 }
@@ -350,14 +344,36 @@ g_closure_unref (GClosure *closure)
   if (closure->ref_count == 0)
     {
       closure_invoke_notifiers (closure, FNOTIFY);
+      g_free (closure->notifiers);
       g_free (closure);
     }
 }
 
 void
-g_closure_remove_inotify (GClosure      *closure,
-			  gpointer       notify_data,
-			  GClosureNotify notify_func)
+g_closure_sink (GClosure *closure)
+{
+  g_return_if_fail (closure != NULL);
+  g_return_if_fail (closure->ref_count > 0);
+
+  /* floating is basically a kludge to avoid creating closures
+   * with a ref_count of 0. so the intial ref_count a closure has
+   * is unowned. with invoking g_closure_sink() code may
+   * indicate that it takes over that intiial ref_count.
+   */
+  if (closure->floating)
+    {
+      closure->floating = FALSE;
+      if (closure->ref_count > 1)
+	closure->ref_count -= 1;
+      else
+	g_closure_unref (closure);
+    }
+}
+
+void
+g_closure_remove_invalidate_notifier (GClosure      *closure,
+				      gpointer       notify_data,
+				      GClosureNotify notify_func)
 {
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
@@ -371,9 +387,9 @@ g_closure_remove_inotify (GClosure      *closure,
 }
 
 void
-g_closure_remove_fnotify (GClosure      *closure,
-                          gpointer       notify_data,
-			  GClosureNotify notify_func)
+g_closure_remove_finalize_notifier (GClosure      *closure,
+				    gpointer       notify_data,
+				    GClosureNotify notify_func)
 {
   g_return_if_fail (closure != NULL);
   g_return_if_fail (notify_func != NULL);
@@ -453,7 +469,7 @@ g_cclosure_new (GCallback      callback_func,
   
   closure = g_closure_new_simple (sizeof (GCClosure), user_data);
   if (destroy_data)
-    g_closure_add_fnotify (closure, user_data, destroy_data);
+    g_closure_add_finalize_notifier (closure, user_data, destroy_data);
   ((GCClosure*) closure)->callback = callback_func;
   
   return closure;
@@ -470,7 +486,7 @@ g_cclosure_new_swap (GCallback      callback_func,
   
   closure = g_closure_new_simple (sizeof (GCClosure), user_data);
   if (destroy_data)
-    g_closure_add_fnotify (closure, user_data, destroy_data);
+    g_closure_add_finalize_notifier (closure, user_data, destroy_data);
   ((GCClosure*) closure)->callback = callback_func;
   closure->derivative_flag = TRUE;
   
@@ -490,7 +506,7 @@ g_type_class_meta_marshal (GClosure       *closure,
   /* GType itype = GPOINTER_TO_UINT (closure->data); */
   guint offset = GPOINTER_TO_UINT (marshal_data);
   
-  class = G_TYPE_INSTANCE_GET_CLASS (g_value_get_as_pointer (param_values + 0), itype, GTypeClass);
+  class = G_TYPE_INSTANCE_GET_CLASS (g_value_peek_pointer (param_values + 0), itype, GTypeClass);
   callback = G_STRUCT_MEMBER (gpointer, class, offset);
   if (callback)
     closure->marshal (closure,
@@ -513,7 +529,7 @@ g_type_iface_meta_marshal (GClosure       *closure,
   GType itype = GPOINTER_TO_UINT (closure->data);
   guint offset = GPOINTER_TO_UINT (marshal_data);
   
-  class = G_TYPE_INSTANCE_GET_INTERFACE (g_value_get_as_pointer (param_values + 0), itype, GTypeClass);
+  class = G_TYPE_INSTANCE_GET_INTERFACE (g_value_peek_pointer (param_values + 0), itype, GTypeClass);
   callback = G_STRUCT_MEMBER (gpointer, class, offset);
   if (callback)
     closure->marshal (closure,
@@ -524,8 +540,8 @@ g_type_iface_meta_marshal (GClosure       *closure,
 }
 
 GClosure*
-g_signal_type_closure_new (GType    itype,
-			   guint    struct_offset)
+g_signal_type_cclosure_new (GType    itype,
+			    guint    struct_offset)
 {
   GClosure *closure;
   
