@@ -28,9 +28,14 @@
  * MT safe
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include "glib.h"
 
 
+#ifndef DISABLE_MEM_POOLS
 struct _GAllocator /* from gmem.c */
 {
   gchar         *name;
@@ -159,7 +164,7 @@ g_slist_free (GSList *list)
 	  last_node->data = NULL;
 	  last_node = last_node->next;
 	}
-      last_node->data = NULL
+      last_node->data = NULL;
 #else /* !ENABLE_GC_FRIENDLY */
       list->data = list->next;  
 #endif /* ENABLE_GC_FRIENDLY */
@@ -189,6 +194,40 @@ g_slist_free_1 (GSList *list)
 {
   _g_slist_free_1 (list);
 }
+#else /* DISABLE_MEM_POOLS */
+
+#define _g_slist_alloc g_slist_alloc
+GSList*
+g_slist_alloc (void)
+{
+  GSList *list;
+  
+  list = g_new0 (GSList, 1);
+  
+  return list;
+}
+
+void
+g_slist_free (GSList *list)
+{
+  GSList *last;
+  
+  while (list)
+    {
+      last = list;
+      list = list->next;
+      g_free (last);
+    }
+}
+
+#define _g_slist_free_1 g_slist_free_1
+void
+g_slist_free_1 (GSList *list)
+{
+  g_free (list);
+}
+
+#endif
 
 GSList*
 g_slist_append (GSList   *list,
@@ -286,29 +325,54 @@ GSList*
 g_slist_remove (GSList        *list,
 		gconstpointer  data)
 {
-  GSList *tmp;
-  GSList *prev;
+  GSList *tmp, *prev = NULL;
 
-  prev = NULL;
   tmp = list;
-
   while (tmp)
     {
       if (tmp->data == data)
 	{
 	  if (prev)
 	    prev->next = tmp->next;
-	  if (list == tmp)
-	    list = list->next;
-
-	  tmp->next = NULL;
-	  g_slist_free (tmp);
+	  else
+	    list = tmp->next;
 
+	  g_slist_free_1 (tmp);
 	  break;
 	}
-
       prev = tmp;
-      tmp = tmp->next;
+      tmp = prev->next;
+    }
+
+  return list;
+}
+
+GSList*
+g_slist_remove_all (GSList        *list,
+		    gconstpointer  data)
+{
+  GSList *tmp, *prev = NULL;
+
+  tmp = list;
+  while (tmp)
+    {
+      if (tmp->data == data)
+	{
+	  GSList *next = tmp->next;
+
+	  if (prev)
+	    prev->next = next;
+	  else
+	    list = next;
+	  
+	  g_slist_free_1 (tmp);
+	  tmp = next;
+	}
+      else
+	{
+	  prev = tmp;
+	  tmp = prev->next;
+	}
     }
 
   return list;
@@ -580,18 +644,26 @@ g_slist_insert_sorted (GSList       *list,
     }
 }
 
-static GSList* 
-g_slist_sort_merge  (GSList      *l1, 
-		     GSList      *l2,
-		     GCompareFunc compare_func)
+static GSList *
+g_slist_sort_merge (GSList   *l1, 
+		    GSList   *l2,
+		    GFunc     compare_func,
+		    gboolean  use_data,
+		    gpointer  user_data)
 {
   GSList list, *l;
+  gint cmp;
 
   l=&list;
 
   while (l1 && l2)
     {
-      if (compare_func(l1->data,l2->data) < 0)
+      if (use_data)
+	cmp = ((GCompareDataFunc) compare_func) (l1->data, l2->data, user_data);
+      else
+	cmp = ((GCompareFunc) compare_func) (l1->data, l2->data);
+
+      if (cmp <= 0)
         {
 	  l=l->next=l1;
 	  l1=l1->next;
@@ -607,9 +679,11 @@ g_slist_sort_merge  (GSList      *l1,
   return list.next;
 }
 
-GSList* 
-g_slist_sort (GSList       *list,
-	      GCompareFunc compare_func)
+static GSList *
+g_slist_sort_real (GSList   *list,
+		   GFunc     compare_func,
+		   gboolean  use_data,
+		   gpointer  user_data)
 {
   GSList *l1, *l2;
 
@@ -630,7 +704,24 @@ g_slist_sort (GSList       *list,
   l2 = l1->next; 
   l1->next = NULL;
 
-  return g_slist_sort_merge (g_slist_sort (list, compare_func),
-			     g_slist_sort (l2,   compare_func),
-			     compare_func);
+  return g_slist_sort_merge (g_slist_sort_real (list, compare_func, use_data, user_data),
+			     g_slist_sort_real (l2, compare_func, use_data, user_data),
+			     compare_func,
+			     use_data,
+			     user_data);
+}
+
+GSList *
+g_slist_sort (GSList       *list,
+	      GCompareFunc  compare_func)
+{
+  return g_slist_sort_real (list, (GFunc) compare_func, FALSE, NULL);
+}
+
+GSList *
+g_slist_sort_with_data (GSList           *list,
+			GCompareDataFunc  compare_func,
+			gpointer          user_data)
+{
+  return g_slist_sort_real (list, (GFunc) compare_func, TRUE, user_data);
 }
