@@ -34,11 +34,7 @@
 #include <sys/select.h>
 #endif /* HAVE_SYS_SELECT_H */
 
-#ifdef _
-#warning "FIXME remove gettext hack"
-#endif
-
-#define _(x) x
+#include "glibintl.h"
 
 static gint g_execute (const gchar  *file,
                        gchar **argv,
@@ -1292,7 +1288,8 @@ g_execute (const gchar *file,
   else
     {
       gboolean got_eacces = 0;
-      char *path, *p, *name, *freeme;
+      const gchar *path, *p;
+      gchar *name, *freeme;
       size_t len;
       size_t pathlen;
 
