@@ -34,8 +34,8 @@
 /* We cheat a bit and cast type values to (char *).  We detect these
    using the &0xff trick.  */
 #define TTYPE(Page, Char) \
-  (((((int) type_table[Page]) & 0xff) == ((int) type_table[Page])) \
-   ? ((int) (type_table[Page])) \
+  (((GPOINTER_TO_INT(type_table[Page]) & 0xff) == GPOINTER_TO_INT(type_table[Page])) \
+   ? GPOINTER_TO_INT(type_table[Page]) \
    : (type_table[Page][Char]))
 
 #define TYPE(Char) (((Char) > (G_UNICODE_LAST_CHAR)) ? G_UNICODE_UNASSIGNED : TTYPE ((Char) >> 8, (Char) & 0xff))
@@ -118,15 +118,24 @@ g_unichar_ispunct (gunichar c)
 gboolean
 g_unichar_isspace (gunichar c)
 {
-  /* special-case these since Unicode thinks they are not spaces */
-  if (c == ' ' || c == '\t' || c == '\n' || c == '\r' ||
-      c == '\f' || c == '\v') /* "the mythical vertical tab" */
-    return TRUE;
-  else
+  switch (c)
     {
-      int t = TYPE (c);
-      return (t == G_UNICODE_SPACE_SEPARATOR || t == G_UNICODE_LINE_SEPARATOR
-              || t == G_UNICODE_PARAGRAPH_SEPARATOR);
+      /* special-case these since Unicode thinks they are not spaces */
+    case '\t':
+    case '\n':
+    case '\r':
+    case '\f':
+    case '\v': /* vertical tab - as if anyone has ever used this... */
+      return TRUE;
+      break;
+      
+    default:
+      {
+        int t = TYPE (c);
+        return (t == G_UNICODE_SPACE_SEPARATOR || t == G_UNICODE_LINE_SEPARATOR
+                || t == G_UNICODE_PARAGRAPH_SEPARATOR);
+      }
+      break;
     }
 }
 
