@@ -64,13 +64,16 @@
 #define G_PATH_LENGTH   2048
 #endif
 
-#ifdef G_OS_WIN32
+#ifdef G_PLATFORM_WIN32
 #  define STRICT			/* Strict typing, please */
 #  include <windows.h>
+#  undef STRICT
 #  include <ctype.h>
+#endif /* G_PLATFORM_WIN32 */
+
+#ifdef G_OS_WIN32
 #  include <direct.h>
-#  include <io.h>
-#endif /* G_OS_WIN32 */
+#endif
 
 #ifdef HAVE_CODESET
 #include <langinfo.h>
@@ -112,7 +115,7 @@ void
 g_atexit (GVoidFunc func)
 {
   gint result;
-  gchar *error = NULL;
+  const gchar *error = NULL;
 
   /* keep this in sync with glib.h */
 
@@ -157,6 +160,52 @@ my_strchrnul (const gchar *str, gchar c)
   return p;
 }
 
+#ifdef G_OS_WIN32
+
+gchar *inner_find_program_in_path (const gchar *program);
+
+gchar*
+g_find_program_in_path (const gchar *program)
+{
+  const gchar *last_dot = strrchr (program, '.');
+
+  if (last_dot == NULL || strchr (last_dot, '\\') != NULL)
+    {
+      const gint program_length = strlen (program);
+      const gchar *pathext = getenv ("PATHEXT");
+      const gchar *p;
+      gchar *decorated_program;
+      gchar *retval;
+
+      if (pathext == NULL)
+	pathext = ".com;.exe;.bat";
+
+      p = pathext;
+      do
+	{
+	  pathext = p;
+	  p = my_strchrnul (pathext, ';');
+
+	  decorated_program = g_malloc (program_length + (p-pathext) + 1);
+	  memcpy (decorated_program, program, program_length);
+	  memcpy (decorated_program+program_length, pathext, p-pathext);
+	  decorated_program [program_length + (p-pathext)] = '\0';
+	  
+	  retval = inner_find_program_in_path (decorated_program);
+	  g_free (decorated_program);
+
+	  if (retval != NULL)
+	    return retval;
+	} while (*p++ != '\0');
+      return NULL;
+    }
+  else
+    return inner_find_program_in_path (program);
+}
+
+#define g_find_program_in_path inner_find_program_in_path
+#endif
+
 /**
  * g_find_program_in_path:
  * @program: a program name
@@ -167,18 +216,37 @@ my_strchrnul (const gchar *str, gchar c)
  * the path. If @program is already an absolute path, returns a copy of
  * @program if @program exists and is executable, and NULL otherwise.
  * 
+ * On Windows, if @program does not have a file type suffix, tries to
+ * append the suffixes in the PATHEXT environment variable (if that
+ * doesn't exists, the suffixes .com, .exe, and .bat) in turn, and
+ * then look for the resulting file name in the same way as
+ * CreateProcess() would. This means first in the directory where the
+ * program was loaded from, then in the current directory, then in the
+ * Windows 32-bit system directory, then in the Windows directory, and
+ * finally in the directories in the PATH environment variable. If
+ * the program is found, the return value contains the full name
+ * including the type suffix.
+ *
  * Return value: absolute path, or NULL
  **/
 gchar*
 g_find_program_in_path (const gchar *program)
 {
-  gchar *path, *p, *name, *freeme;
+  const gchar *path, *p;
+  gchar *name, *freeme;
+#ifdef G_OS_WIN32
+  gchar *path_tmp;
+#endif
   size_t len;
   size_t pathlen;
 
   g_return_val_if_fail (program != NULL, NULL);
 
-  if (*program == '/')
+  /* If it is an absolute path, or a relative path including subdirectories,
+   * don't look in PATH.
+   */
+  if (g_path_is_absolute (program)
+      || strchr (program, G_DIR_SEPARATOR) != NULL)
     {
       if (g_file_test (program, G_FILE_TEST_IS_EXECUTABLE))
         return g_strdup (program);
@@ -187,6 +255,7 @@ g_find_program_in_path (const gchar *program)
     }
   
   path = g_getenv ("PATH");
+#ifdef G_OS_UNIX
   if (path == NULL)
     {
       /* There is no `PATH' in the environment.  The default
@@ -201,6 +270,23 @@ g_find_program_in_path (const gchar *program)
       
       path = "/bin:/usr/bin:.";
     }
+#else
+  {
+    gchar *tmp;
+    gchar moddir[MAXPATHLEN], sysdir[MAXPATHLEN], windir[MAXPATHLEN];
+
+    GetModuleFileName (NULL, moddir, sizeof (moddir));
+    tmp = g_path_get_dirname (moddir);
+    GetSystemDirectory (sysdir, sizeof (sysdir));
+    GetWindowsDirectory (windir, sizeof (windir));
+    path_tmp = g_strconcat (tmp, ";.;", sysdir, ";", windir,
+			    (path != NULL ? ";" : NULL),
+			    (path != NULL ? path : NULL),
+			    NULL);
+    g_free (tmp);
+    path = path_tmp;
+  }
+#endif
   
   len = strlen (program) + 1;
   pathlen = strlen (path);
@@ -210,7 +296,7 @@ g_find_program_in_path (const gchar *program)
   memcpy (name + pathlen + 1, program, len);
   name = name + pathlen;
   /* And add the slash before the filename  */
-  *name = '/';
+  *name = G_DIR_SEPARATOR;
   
   p = path;
   do
@@ -218,7 +304,7 @@ g_find_program_in_path (const gchar *program)
       char *startp;
 
       path = p;
-      p = my_strchrnul (path, ':');
+      p = my_strchrnul (path, G_SEARCHPATH_SEPARATOR);
 
       if (p == path)
         /* Two adjacent colons, or a colon at the beginning or the end
@@ -233,12 +319,18 @@ g_find_program_in_path (const gchar *program)
           gchar *ret;
           ret = g_strdup (startp);
           g_free (freeme);
+#ifdef G_OS_WIN32
+	  g_free (path_tmp);
+#endif
           return ret;
         }
     }
   while (*p++ != '\0');
   
   g_free (freeme);
+#ifdef G_OS_WIN32
+  g_free (path_tmp);
+#endif
 
   return NULL;
 }
@@ -374,18 +466,18 @@ g_parse_debug_string  (const gchar *string,
   return result;
 }
 
-gchar*
+G_CONST_RETURN gchar*
 g_basename (const gchar	   *file_name)
 {
   register gchar *base;
-#ifdef G_ENABLE_DEBUG
+#if defined(G_ENABLE_DEBUG) && !defined(G_OS_WIN32)
   static gboolean first_call = TRUE;
 
   if (first_call)
     {
-      g_warning("g_basename is deprecated. Use g_path_get_basename instead.");
-      g_warning("Watch out! You have to g_free the string returned by "
-		"g_path_get_basename.");
+      g_message ("g_basename is deprecated. Use g_path_get_basename instead. "
+		 "Beware that the string returned by g_path_get_basename() has "
+		 " to be g_free()ed.");
       first_call = FALSE;
     }
 #endif /* G_ENABLE_DEBUG */
@@ -459,24 +551,56 @@ g_path_is_absolute (const gchar *file_name)
     return TRUE;
 
 #ifdef G_OS_WIN32
+  /* Recognize drive letter on native Windows */
   if (isalpha (file_name[0]) && file_name[1] == ':' && file_name[2] == G_DIR_SEPARATOR)
     return TRUE;
-#endif
+#endif /* G_OS_WIN32 */
 
   return FALSE;
 }
 
-gchar*
-g_path_skip_root (gchar *file_name)
+G_CONST_RETURN gchar*
+g_path_skip_root (const gchar *file_name)
 {
   g_return_val_if_fail (file_name != NULL, NULL);
   
+#ifdef G_PLATFORM_WIN32
+  /* Skip \\server\share (Win32) or //server/share (Cygwin) */
+  if (file_name[0] == G_DIR_SEPARATOR &&
+      file_name[1] == G_DIR_SEPARATOR &&
+      file_name[2])
+    {
+      gchar *p;
+
+      if ((p = strchr (file_name + 2, G_DIR_SEPARATOR)) > file_name + 2 &&
+	  p[1])
+	{
+	  file_name = p + 1;
+
+	  while (file_name[0] && file_name[0] != G_DIR_SEPARATOR)
+	    file_name++;
+
+	  /* Possibly skip a backslash after the share name */
+	  if (file_name[0] == G_DIR_SEPARATOR)
+	    file_name++;
+
+	  return (gchar *)file_name;
+	}
+    }
+#endif
+  
+  /* Skip initial slashes */
   if (file_name[0] == G_DIR_SEPARATOR)
-    return file_name + 1;
+    {
+      while (file_name[0] == G_DIR_SEPARATOR)
+	file_name++;
+      return (gchar *)file_name;
+    }
 
 #ifdef G_OS_WIN32
+  /* Skip X:\ */
   if (isalpha (file_name[0]) && file_name[1] == ':' && file_name[2] == G_DIR_SEPARATOR)
-    return file_name + 3;
+    return (gchar *)file_name + 3;
 #endif
 
   return NULL;
@@ -507,12 +631,12 @@ g_path_get_dirname (const gchar	   *file_name)
 gchar*
 g_dirname (const gchar	   *file_name)
 {
-#ifdef G_ENABLE_DEBUG
+#if defined(G_ENABLE_DEBUG) && !defined(G_OS_WIN32)
   static gboolean first_call = TRUE;
 
   if (first_call)
     {
-      g_warning("g_dirname is deprecated. Use g_path_get_dirname instead.");
+      g_message ("g_dirname() is deprecated. Use g_path_get_dirname() instead.");
       first_call = FALSE;
     }
 #endif /* G_ENABLE_DEBUG */
@@ -567,7 +691,7 @@ g_get_current_dir (void)
   return dir;
 }
 
-gchar*
+G_CONST_RETURN gchar*
 g_getenv (const gchar *variable)
 {
 #ifndef G_OS_WIN32
@@ -687,15 +811,30 @@ g_get_any_init (void)
 	g_home_dir = g_strdup (g_getenv ("HOME"));
       
 #ifdef G_OS_WIN32
+      /* In case HOME is Unix-style (it happens), convert it to
+       * Windows style.
+       */
+      if (g_home_dir)
+	{
+	  gchar *p;
+	  while ((p = strchr (g_home_dir, '/')) != NULL)
+	    *p = '\\';
+	}
+
       if (!g_home_dir)
 	{
-	  /* The official way to specify a home directory on NT is
-	   * the HOMEDRIVE and HOMEPATH environment variables.
-	   *
-	   * This is inside #ifdef G_OS_WIN32 because with the cygwin dll,
-	   * HOME should be a POSIX style pathname.
+	  /* USERPROFILE is probably the closest equivalent to $HOME? */
+	  if (getenv ("USERPROFILE") != NULL)
+	    g_home_dir = g_strdup (g_getenv ("USERPROFILE"));
+	}
+
+      if (!g_home_dir)
+	{
+	  /* At least at some time, HOMEDRIVE and HOMEPATH were used
+	   * to point to the home directory, I think. But on Windows
+	   * 2000 HOMEDRIVE seems to be equal to SYSTEMDRIVE, and
+	   * HOMEPATH is its root "\"?
 	   */
-	  
 	  if (getenv ("HOMEDRIVE") != NULL && getenv ("HOMEPATH") != NULL)
 	    {
 	      gchar *homedrive, *homepath;
@@ -828,7 +967,7 @@ g_get_any_init (void)
     }
 }
 
-gchar*
+G_CONST_RETURN gchar*
 g_get_user_name (void)
 {
   G_LOCK (g_utils_global);
@@ -839,7 +978,7 @@ g_get_user_name (void)
   return g_user_name;
 }
 
-gchar*
+G_CONST_RETURN gchar*
 g_get_real_name (void)
 {
   G_LOCK (g_utils_global);
@@ -856,7 +995,7 @@ g_get_real_name (void)
  * deduced, return NULL.
  */
 
-gchar*
+G_CONST_RETURN gchar*
 g_get_home_dir (void)
 {
   G_LOCK (g_utils_global);
@@ -874,7 +1013,7 @@ g_get_home_dir (void)
  * and C:\ on Windows.
  */
 
-gchar*
+G_CONST_RETURN gchar*
 g_get_tmp_dir (void)
 {
   G_LOCK (g_utils_global);
@@ -952,15 +1091,48 @@ g_get_codeset (void)
   char *result = nl_langinfo (CODESET);
   return g_strdup (result);
 #else
-#ifndef G_OS_WIN32
+#ifdef G_PLATFORM_WIN32
+  return g_strdup_printf ("CP%d", GetACP ());
+#else
   /* FIXME: Do something more intelligent based on setlocale (LC_CTYPE, NULL)
    */
   return g_strdup ("ISO-8859-1");
-#else
-  /* On Win32 we always use UTF-8. At least in GDK. SO should we
-   * therefore return that?
-   */
-  return g_strdup ("UTF-8");
 #endif
 #endif
 }
+
+#ifdef ENABLE_NLS
+
+#include <libintl.h>
+
+
+#ifndef GLIB_LOCALE_DIR
+#ifdef G_PLATFORM_WIN32
+
+#define GLIB_LOCALE_DIR					       	\
+  g_win32_get_package_installation_subdirectory			\
+  (GETTEXT_PACKAGE, g_strdup_printf ("glib-%d.%d.dll",		\
+				     GLIB_MAJOR_VERSION,	\
+				     GLIB_MINOR_VERSION),	\
+   "locale")
+
+#endif /* G_PLATFORM_WIN32 */
+#endif /* !GLIB_LOCALE_DIR */
+
+G_CONST_RETURN gchar *
+_glib_gettext (const gchar *str)
+{
+  gboolean _glib_gettext_initialized = FALSE;
+
+  if (!_glib_gettext_initialized)
+    {
+      bindtextdomain(GETTEXT_PACKAGE, GLIB_LOCALE_DIR);
+      _glib_gettext_initialized = TRUE;
+    }
+  
+  return dgettext (GETTEXT_PACKAGE, str);
+}
+
+#endif /* ENABLE_NLS */
+
+
