@@ -42,7 +42,10 @@
 
 #define STRICT			/* Strict typing, please */
 #include <windows.h>
+#undef STRICT
+#ifndef G_WITH_CYGWIN
 #include <direct.h>
+#endif
 #include <errno.h>
 #include <ctype.h>
 #ifdef _MSC_VER
@@ -51,7 +54,13 @@
 
 #include "glib.h"
 
-int
+#ifdef G_WITH_CYGWIN
+#include <sys/cygwin.h>
+#endif
+
+#ifndef G_WITH_CYGWIN
+
+gint
 g_win32_ftruncate (gint  fd,
 		   guint size)
 {
@@ -214,8 +223,9 @@ g_win32_closedir (DIR *dir)
 
   return 0;
 }
+#endif
 
-/* Mingw32 headers don't have latest language and sublanguage codes */
+/* Mingw headers don't have latest language and sublanguage codes */
 #ifndef LANG_AFRIKAANS
 #define LANG_AFRIKAANS 0x36
 #endif
@@ -526,6 +536,17 @@ g_win32_closedir (DIR *dir)
 #define SUBLANG_UZBEK_CYRILLIC 0x02
 #endif
 
+/**
+ * g_win32_getlocale:
+ *
+ * The setlocale in the Microsoft C library uses locale names of the
+ * form "English_United States.1252" etc. We want the Unixish standard
+ * form "en", "zh_TW" etc. This function gets the current thread
+ * locale from Windows and returns it as a string of the above form
+ * for use in forming file names etc. The returned string should be
+ * deallocated with g_free().
+ */
+
 gchar *
 g_win32_getlocale (void)
 {
@@ -625,7 +646,7 @@ g_win32_getlocale (void)
 	case SUBLANG_ENGLISH_CAN: sl = "CA"; break;
 	case SUBLANG_ENGLISH_NZ: sl = "NZ"; break;
 	case SUBLANG_ENGLISH_EIRE: sl = "IE"; break;
-	case SUBLANG_ENGLISH_SOUTH_AFRICA: sl = "SA"; break;
+	case SUBLANG_ENGLISH_SOUTH_AFRICA: sl = "ZA"; break;
 	case SUBLANG_ENGLISH_JAMAICA: sl = "JM"; break;
 	case SUBLANG_ENGLISH_CARIBBEAN: sl = "@caribbean"; break; /* ??? */
 	case SUBLANG_ENGLISH_BELIZE: sl = "BZ"; break;
@@ -711,7 +732,7 @@ g_win32_getlocale (void)
       switch (sub)
 	{
 	/* SUBLANG_NORWEGIAN_BOKMAL == SUBLANG_DEFAULT */
-	case SUBLANG_NORWEGIAN_NYNORSK: sl = "@nynorsk"; break;
+	case SUBLANG_NORWEGIAN_NYNORSK: l = "nn"; break;
 	}
       break;
     case LANG_ORIYA: l = "or"; break;
@@ -791,13 +812,24 @@ g_win32_getlocale (void)
   strcpy (bfr, l);
   if (sl != NULL)
     {
-      strcat (bfr, "_");
+      if (sl[0] != '@')
+	strcat (bfr, "_");
       strcat (bfr, sl);
     }
 
   return g_strdup (bfr);
 }
 
+/**
+ * g_win32_error_message:
+ * @error: error code
+ *
+ * Translate a Win32 error code (as returned by GetLastError()) into
+ * the corresponding message. The message is either language neutral,
+ * or in the thread's language, or the user's language, the system's
+ * langauge, or US English (see docs for FormatMessage). The returned
+ * string should be deallocated with g_free().
+ */
 gchar *
 g_win32_error_message (gint error)
 {
@@ -822,3 +854,175 @@ g_win32_error_message (gint error)
 
   return retval;
 }
+
+static gchar *
+get_package_directory_from_module (gchar *module_name)
+{
+  static GHashTable *module_dirs = NULL;
+  HMODULE hmodule = NULL;
+  gchar *fn;
+  gchar *p;
+  gchar *result;
+
+  if (module_dirs == NULL)
+    module_dirs = g_hash_table_new (g_str_hash, g_str_equal);
+  
+  result = g_hash_table_lookup (module_dirs, module_name ? module_name : "");
+      
+  if (result)
+    return g_strdup (result);
+
+  if (module_name)
+    {
+      hmodule = GetModuleHandle (module_name);
+      if (!hmodule)
+	return NULL;
+    }
+
+  fn = g_malloc (MAX_PATH);
+  if (!GetModuleFileName (hmodule, fn, MAX_PATH))
+    return NULL;
+
+#ifdef G_WITH_CYGWIN
+  /* In Cygwin we need to have POSIX paths */
+  {
+    gchar tmp[MAX_PATH];
+
+    cygwin_conv_to_posix_path(fn, tmp);
+    g_free(fn);
+    fn = g_strdup(tmp);
+  }
+#endif
+
+  if ((p = strrchr (fn, G_DIR_SEPARATOR)) != NULL)
+    *p = '\0';
+
+  if (module_name)
+    {
+      p = strrchr (fn, G_DIR_SEPARATOR);
+      if (p && (g_strcasecmp (p + 1, "bin") == 0 ||
+		g_strcasecmp (p + 1, "lib") == 0))
+	*p = '\0';
+    }
+
+  g_hash_table_insert (module_dirs, module_name ? module_name : "", fn);
+
+  return g_strdup (fn);
+}
+
+/**
+ * g_win32_get_package_installation_directory:
+ * @package: An identifier for a software package, or NULL
+ * @dll_name: The name of a DLL that a package provides, or NULL
+ *
+ * Try to determine the installation directory for a software package.
+ * Typically used by GNU software packages.
+ *
+ * @package should be a short identifier for the package. Typically it
+ * is the same identifier as used for GETTEXT_PACKAGE in software
+ * configured accoring to GNU standards. The function first looks in
+ * the Windows Registry for the value #InstallationDirectory in the
+ * key #HKLM\Software\@package, and if that value exists and is a
+ * string, returns that.
+ *
+ * If @package is NULL, or the above value isn't found in the
+ * Registry, but @dll_name is non-NULL, it should name a DLL loaded
+ * into the current process. Typically that would be the name of the
+ * DLL calling this function, looking for its installation
+ * directory. The function then asks Windows what directory that DLL
+ * was loaded from. If that directory's last component is "bin" or
+ * "lib", the parent directory is returned, otherwise the directory
+ * itself. If that DLL isn't loaded, the function proceeds as if
+ * @dll_name was NULL.
+ *
+ * If both @package and @dll_name are NULL, the directory from where
+ * the main executable of the process was loaded is uses instead in
+ * the same way as above.
+ *
+ * The return value should be freed with g_free() when not needed any longer.  */
+
+gchar *
+g_win32_get_package_installation_directory (gchar *package,
+					    gchar *dll_name)
+{
+  static GHashTable *package_dirs = NULL;
+  gchar *result = NULL;
+  gchar *key;
+  HKEY reg_key = NULL;
+  DWORD type;
+  DWORD nbytes;
+
+  if (package != NULL)
+    {
+      if (package_dirs == NULL)
+	package_dirs = g_hash_table_new (g_str_hash, g_str_equal);
+      
+      result = g_hash_table_lookup (package_dirs, package);
+      
+      if (result && result[0])
+	return g_strdup (result);
+      
+      key = g_strconcat ("Software\\", package, NULL);
+      
+      nbytes = 0;
+      if (RegOpenKeyEx (HKEY_LOCAL_MACHINE, key, 0,
+			KEY_QUERY_VALUE, &reg_key) == ERROR_SUCCESS
+	  && RegQueryValueEx (reg_key, "InstallationDirectory", 0,
+			      &type, NULL, &nbytes) == ERROR_SUCCESS
+	  && type == REG_SZ)
+	{
+	  result = g_malloc (nbytes + 1);
+	  RegQueryValueEx (reg_key, "InstallationDirectory", 0,
+			   &type, result, &nbytes);
+	  result[nbytes] = '\0';
+	}
+
+      if (reg_key != NULL)
+	RegCloseKey (reg_key);
+      
+      g_free (key);
+      
+    }
+  if (result)
+    {
+      g_hash_table_insert (package_dirs, package, result);
+      return g_strdup (result);
+    }
+
+  if (dll_name != NULL)
+    result = get_package_directory_from_module (dll_name);
+
+  if (result == NULL)
+    result = get_package_directory_from_module (NULL);
+
+  return result;
+}
+
+/**
+ * g_win32_get_package_installation_subdirectory:
+ * @package: An identifier for a software package, or NULL
+ * @dll_name: The name of a DLL that a package provides, or NULL
+ * @subdir: A subdirectory of the package installation directory.
+ *
+ * Returns a string containg the path of the subdirectory @subdir in
+ * the return value from calling
+ * g_win32_get_package_installation_directory() with the @package and
+ * @dll_name parameters. The return value should be freed with
+ * g_free() when no longer needed.
+ */
+
+gchar *
+g_win32_get_package_installation_subdirectory (gchar *package,
+					       gchar *dll_name,
+					       gchar *subdir)
+{
+  gchar *prefix;
+  gchar *sep;
+
+  prefix = g_win32_get_package_installation_directory (package, dll_name);
+
+  sep = (prefix[strlen (prefix) - 1] == G_DIR_SEPARATOR ?
+	 "" : G_DIR_SEPARATOR_S);
+
+  return g_strconcat (prefix, sep, subdir, NULL);
+}
