@@ -28,6 +28,10 @@
 
 #undef G_LOG_DOMAIN
 
+#ifdef GLIB_COMPILATION
+#undef GLIB_COMPILATION
+#endif
+
 #include <stdio.h>
 #include <string.h>
 #include <errno.h>
@@ -42,8 +46,8 @@
 #include <io.h>			/* For read(), write() etc */
 #endif
 
-int array[10000];
-gboolean failed = FALSE;
+static int array[10000];
+static gboolean failed = FALSE;
 
 #define	TEST(m,cond)	G_STMT_START { failed = !(cond); \
 if (failed) \
@@ -333,28 +337,54 @@ main (int   argc,
     { "/", "/" },
     { "////", "/" },
     { ".////", "." },
-    { ".", "." },
-    { "..", "." },
     { "../", ".." },
     { "..////", ".." },
-    { "", "." },
     { "a/b", "a" },
     { "a/b/", "a/b" },
     { "c///", "c" },
 #else
     { "\\", "\\" },
     { ".\\\\\\\\", "." },
-    { ".", "." },
-    { "..", "." },
     { "..\\", ".." },
     { "..\\\\\\\\", ".." },
-    { "", "." },
     { "a\\b", "a" },
     { "a\\b\\", "a\\b" },
     { "c\\\\\\", "c" },
 #endif
+#ifdef G_WITH_CYGWIN
+    { "//server/share///x", "//server/share" },
+#endif
+    { ".", "." },
+    { "..", "." },
+    { "", "." },
   };
-  guint n_dirname_checks = sizeof (dirname_checks) / sizeof (dirname_checks[0]);
+  guint n_dirname_checks = G_N_ELEMENTS (dirname_checks);
+
+  struct {
+    gchar *filename;
+    gchar *without_root;
+  } skip_root_checks[] = {
+#ifndef G_OS_WIN32
+    { "/", "" },
+    { "//", "" },
+    { "/foo", "foo" },
+    { "//foo", "foo" },
+    { "a/b", NULL },
+#else
+    { "\\", "" },
+    { "\\foo", "foo" },
+    { "\\\\server\\foo", "" },
+    { "\\\\server\\foo\\bar", "bar" },
+    { "a\\b", NULL },
+#endif
+#ifdef G_WITH_CYGWIN
+    { "//server/share///x", "//x" },
+#endif
+    { ".", NULL },
+    { "", NULL },
+  };
+  guint n_skip_root_checks = G_N_ELEMENTS (skip_root_checks);
+
   guint16 gu16t1 = 0x44afU, gu16t2 = 0xaf44U;
   guint32 gu32t1 = 0x02a7f109U, gu32t2 = 0x09f1a702U;
 #ifdef G_HAVE_GINT64
@@ -367,6 +397,17 @@ main (int   argc,
   char template[10];
   GError *error;
   char *name_used;
+  gchar *p;
+#ifdef G_OS_WIN32
+  gchar *glib_dll = g_strdup_printf ("glib-%d.%d.dll",
+				     GLIB_MAJOR_VERSION,
+				     GLIB_MINOR_VERSION);
+#endif
+#ifdef G_WITH_CYGWIN
+  gchar *glib_dll = g_strdup_printf ("cygglib-%d.%d.dll",
+				     GLIB_MAJOR_VERSION,
+				     GLIB_MINOR_VERSION);
+#endif
 
   g_print ("TestGLib v%u.%u.%u (i:%u b:%u)\n",
 	   glib_major_version,
@@ -416,6 +457,28 @@ main (int   argc,
   if (n_dirname_checks)
     g_print ("ok\n");
 
+  g_print ("checking g_path_skip_root()...");
+  for (i = 0; i < n_skip_root_checks; i++)
+    {
+      const gchar *skipped;
+
+      skipped = g_path_skip_root (skip_root_checks[i].filename);
+      if ((skipped && !skip_root_checks[i].without_root) ||
+	  (!skipped && skip_root_checks[i].without_root) ||
+	  ((skipped && skip_root_checks[i].without_root) &&
+	   strcmp (skipped, skip_root_checks[i].without_root)))
+	{
+	  g_print ("\nfailed for \"%s\"==\"%s\" (returned: \"%s\")\n",
+		   skip_root_checks[i].filename,
+		   (skip_root_checks[i].without_root ?
+		    skip_root_checks[i].without_root : "<NULL>"),
+		   (skipped ? skipped : "<NULL>"));
+	  n_skip_root_checks = 0;
+	}
+    }
+  if (n_skip_root_checks)
+    g_print ("ok\n");
+
   g_print ("checking doubly linked lists...");
 
   list = NULL;
@@ -1099,8 +1162,24 @@ main (int   argc,
 
   g_print ("ok\n");
 
-#ifdef G_OS_WIN32
+#ifdef G_PLATFORM_WIN32
   g_print ("current locale: %s\n", g_win32_getlocale ());
+
+  g_print ("GLib installation directory, from Registry entry for %s if available: %s\n",
+	   GETTEXT_PACKAGE,
+	   g_win32_get_package_installation_directory (GETTEXT_PACKAGE, NULL));
+  g_print ("Ditto, or from GLib DLL name: %s\n",
+	   g_win32_get_package_installation_directory (GETTEXT_PACKAGE, glib_dll));
+  g_print ("Ditto, only from GLib DLL name: %s\n",
+	   g_win32_get_package_installation_directory (NULL, glib_dll));
+  g_print ("locale subdirectory of GLib installation directory: %s\n",
+	   g_win32_get_package_installation_subdirectory (NULL, glib_dll, "locale"));
+  g_print ("GTK+ 2.0 installation directory, if available: %s\n",
+	   g_win32_get_package_installation_directory ("gtk20", NULL));
+
+  g_print ("found more.com as %s\n", g_find_program_in_path ("more.com"));
+  g_print ("found regedit as %s\n", g_find_program_in_path ("regedit"));
+
 #endif
 
   g_print ("checking file functions...\n");
