@@ -3,20 +3,28 @@
  * Copyright (C) 1999 The Free Software Foundation
  *
  * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Library General Public
+ * modify it under the terms of the GNU Lesser General Public
  * License as published by the Free Software Foundation; either
  * version 2 of the License, or (at your option) any later version.
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
- * Library General Public License for more details.
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Library General Public
+ * You should have received a copy of the GNU Lesser General Public
  * License along with this library; if not, write to the
  * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  * Boston, MA 02111-1307, USA.
  */
+
+/*
+ * Modified by the GLib Team and others 1997-2000.  See the AUTHORS
+ * file for a list of people on the GLib Team.  See the ChangeLog
+ * files for a list of changes.  These files are distributed with
+ * GLib at ftp://ftp.gtk.org/pub/gtk/. 
+ */
+
 #undef G_LOG_DOMAIN
 
 #ifdef HAVE_CONFIG_H
@@ -76,9 +84,9 @@ my_hash (gconstpointer key)
   return (guint) *((const gint*) key);
 }
 
-static gint
-my_hash_compare (gconstpointer a,
-		 gconstpointer b)
+static gboolean
+my_hash_equal (gconstpointer a,
+	       gconstpointer b)
 {
   return *((const gint*) a) == *((const gint*) b);
 }
@@ -152,11 +160,9 @@ static guint honeyman_hash(gconstpointer key)
 }
 
 
-static gint second_hash_cmp (gconstpointer a, gconstpointer b)
+static gboolean second_hash_cmp (gconstpointer a, gconstpointer b)
 {
-  gint rc = (strcmp (a, b) == 0);
-
-  return rc;
+  return (strcmp (a, b) == 0);
 }
 
 
@@ -328,7 +334,7 @@ main (int   argc,
   GHashTable *hash_table;
   gint i;
 
-  hash_table = g_hash_table_new (my_hash, my_hash_compare);
+  hash_table = g_hash_table_new (my_hash, my_hash_equal);
   for (i = 0; i < 10000; i++)
     {
       array[i] = i;
